#include <LiquidCrystal_I2C.h>

// USB 键盘解析库
#include "USBReader.h"
// 大数运算库
#include "BigNumber.h"
#include "Keyboard.h"

typedef struct Number {
  char data[8];
};

// 语音的数据
Number VoiceData[19] =
{
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x96, 0xF6, 0x9A, }, //零 0
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x4E, 0x0, 0xB4,  }, //一 1
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x4E, 0x8C, 0x38, }, //二 2
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x4E, 0x9, 0xBD,  }, //三 3
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x56, 0xDB, 0x77, }, //四 4
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x4E, 0x94, 0x20, }, //五 5
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x51, 0x6D, 0xC6, }, //六 6
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x4E, 0x3, 0xB7,  }, //七 7
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x51, 0x6B, 0xC0, }, //八 8
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x4E, 0x5D, 0xE9, }, //九 9
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x52, 0xA0, 0x8,  }, //+  10
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x51, 0xCF, 0x64, }, //-  11
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x4E, 0x58, 0xEC, }, //*  12
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x96, 0x64, 0x8   }, ///  13
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x5F, 0x52, 0xF7, }, //归 14
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x7B, 0x49, 0xC8, }, //等 15
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x4E, 0x8E, 0x3A, }, //于 16
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x8D, 0x1F, 0x68, }, //负 17
  {0xFD, 0x0, 0x5, 0x1, 0x3, 0x70, 0xB9, 0x33, }
};//点 18

// Max length of a number
#define MAXLENGTH 10
String  Num1;
String  Num2;
char    Operator;

// 当前状态 0 to 6
byte    CurrentState = 0;
byte    RecordStatus = 0xFF;

// 从键盘获得的字符
char    c = 0;

KeyReport  NullData = {0, 0, 0, 0, 0, 0, 0, 0};

USBReader usb_reader(&on_usb_data);

LiquidCrystal_I2C lcd(0x27, 16, 2);

//
// 将键盘 ScanCode 转化为字符
//
void on_usb_data(USBKeyboardData *data) {
  switch (data->Data[2]) {
    case 0x53: // 用 N 表示归零
      c = 'N';
      break;
    case 0x54:
      c = '/';
      break;
    case 0x55:
      c = '*';
      break;
    case 0x56:
      c = '-';
      break;
    case 0x57:
      c = '+';
      break;
    case 0x58: // E 表示回车
      c = 'E';
      break;
    case 0x63:
      c = '.';
      break;
    case 0x62:
      c = '0';
      break;
    case 0x59:
      c = '1';
      break;
    case 0x5A:
      c = '2';
      break;
    case 0x5B:
      c = '3';
      break;
    case 0x5C:
      c = '4';
      break;
    case 0x5D:
      c = '5';
      break;
    case 0x5E:
      c = '6';
      break;
    case 0x5F:
      c = '7';
      break;
    case 0x60:
      c = '8';
      break;
    case 0x61:
      c = '9';
      break;
    default:
      c = 0;
  }
}

//
// 大数运算
//
BigNumber Calculate(String a, String b, char op) {

  // Convert a to BugNumber x
  char xArr[a.length() + 1];
  a.toCharArray(xArr, a.length() + 1);
  BigNumber x (xArr);

  //  Convert b to BugNumber y
  char yArr[b.length() + 1];
  b.toCharArray(yArr, b.length() + 1);
  BigNumber y (yArr);

  if (op == '+') {
    x += y;
  }

  if (op == '-') {
    x -= y;
  }
  if (op == '*') {
    x *= y;
  }
  if (op == '/') {
    x /=  y;
  }
#ifdef KB_DEBUG
  Serial.print(F("Result [")); Serial.print(x); Serial.println("]");
#endif

  //Return result in x
  return x;
}

//
// Print current status and number for debug
//
void PrintStatus() {
#ifdef KB_DEBUG
  Serial.print(F("Status:"));
  Serial.print(CurrentState);
  Serial.print(F("  Num1["));
  Serial.print(Num1);
  Serial.print(F("] Op["));
  Serial.print(Operator);
  Serial.print(F("]  Num2["));
  Serial.print(Num2);
  Serial.println(F("]"));
#endif

}
void LCDShow(String str)
{
  lcd.clear();
  lcd.setCursor(0, 0);
  if (str.length() <= 16) {
    lcd.print(str);
  }
  else {
    for (int i = 0; i < 16; i++) {
      lcd.print(str[i]);
    }
    lcd.setCursor(0, 1);
    for (int i = 16; i < str.length(); i++) {
      lcd.print(str[i]);
    }
  }
}
//
// 检查当前输入的字符串是否超过上限
//
boolean ValidInput(String s) {
  if (s.length() + 1 == MAXLENGTH) {
    return false;
  }
  return true;
}

void setup()
{
  // 设置大数最大长度
  BigNumber::begin (MAXLENGTH);

#ifdef KB_DEBUG
  Serial.begin(115200);
#endif

  Serial1.begin(9600);

  usb_reader.setup();

  Keyboard.begin();

  lcd.init();

  // Print a message to the LCD.
  lcd.backlight();
  lcd.setCursor(0, 0);
  lcd.print("KB Calculator");
}

// 语音输出字符
void SpeakOut(uint8_t chr) {
  switch (chr) {
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      {
        for (int j = 0; j < 8; j++) {
          Serial1.write(VoiceData[chr - '0'].data[j]);
        }
        break;
      }
    case '+':
      {
        for (int j = 0; j < 8; j++) {
          Serial1.write(VoiceData[10].data[j]);
        }
        break;
      }
    case '-':
      {
        for (int j = 0; j < 8; j++) {
          Serial1.write(VoiceData[11].data[j]);
        }
        break;
      }
    case '*':
      {
        for (int j = 0; j < 8; j++) {
          Serial1.write(VoiceData[12].data[j]);
        }
        break;
      }
    case '/':
      {
        for (int j = 0; j < 8; j++) {
          Serial1.write(VoiceData[13].data[j]);
        }
        break;
      }
    case 'F':  // 负
      {
        for (int j = 0; j < 8; j++) {
          Serial1.write(VoiceData[17].data[j]);
        }
        delay(500);
        break;
      }
    case 'N': // 归零
      {
        // "归"
        for (int j = 0; j < 8; j++) {
          Serial1.write(VoiceData[14].data[j]);
        }
        delay(250);
        // "零"
        for (int j = 0; j < 8; j++) {
          Serial1.write(VoiceData[0].data[j]);
        }
        break;
      }
    case 'E': // 回车==等于
      {
        for (int j = 0; j < 8; j++) {
          Serial1.write(VoiceData[15].data[j]);
        }
        delay(300);
        for (int j = 0; j < 8; j++) {
          Serial1.write(VoiceData[16].data[j]);
        }
        // 因为需要语音输出，所以延时很长对于回车来说需要及时发送全0来防止出现一直输入回车的情况
        Keyboard.sendReport(&NullData);

        break;
      }
    case '.':
      {
        for (int j = 0; j < 8; j++) {
          Serial1.write(VoiceData[18].data[j]);
        }
        break;
      }
    default:
      break;
  }
  delay(400);
}

void loop()
{
  usb_reader.loop();

  // Start when input is not empty
  if (c != 0) {
    switch (CurrentState) {
      case 0:
        if ((c >= '0') && (c <= '9')) {
          Num1 = c;
          Num2 = "";
          Operator = ' ';
          PrintStatus();
          CurrentState = 1;
          SpeakOut(c);
        }
        break;
      case 1:
        if ((c >= '0') && (c <= '9')) {
          if (ValidInput(Num1)) {
            Num1 = Num1 + c;
            SpeakOut(c);
            PrintStatus();
          }
        }
        if (c == '.') {
          if (ValidInput(Num1)) {
            Num1 = Num1 + c;
            CurrentState = 2;
            PrintStatus();
            SpeakOut(c);
          }
        }
        if ((c == '+') || (c == '-') || (c == '*') || (c == '/')) {
          Operator = c;
          CurrentState = 3;
          PrintStatus();
          SpeakOut(c);
        }
        break;
      case 2:
        if ((c >= '0') && (c <= '9')) {
          if (ValidInput(Num1)) {
            Num1 = Num1 + c;
            CurrentState = 21;
            PrintStatus();
            SpeakOut(c);
          }
        }
        break;
      case 21:
        if ((c >= '0') && (c <= '9')) {
          if (ValidInput(Num1)) {
            Num1 = Num1 + c;
            PrintStatus();
            SpeakOut(c);
          }
        }
        if ((c == '+') || (c == '-') || (c == '*') || (c == '/')) {
          Operator = c;
          CurrentState = 3;
          PrintStatus();
          SpeakOut(c);
        }

        break;
      case 3:
        if ((c == '+') || (c == '-') || (c == '*') || (c == '/')) {
          Operator = c;
          PrintStatus();
          SpeakOut(c);
        }
        if ((c >= '0') && (c <= '9')) {
          if (ValidInput(Num2)) {
            Num2 = Num2 + c;
            CurrentState = 4;
            PrintStatus();
            SpeakOut(c);
          }
        }
        break;
      case 4:
        if ((c >= '0') && (c <= '9')) {
          if (ValidInput(Num2)) {
            Num2 = Num2 + c;
            PrintStatus();
            SpeakOut(c);
          }
        }
        if (c == '.') {
          if (ValidInput(Num2)) {
            Num2 = Num2 + c;
            CurrentState = 5;
            PrintStatus();
            SpeakOut(c);
          }
        }
        if (c == 'E') {
          Num1 = Calculate(Num1, Num2, Operator).toString();
          CurrentState = 6;
          PrintStatus();
          SpeakOut(c);
        }

        break;
      case 5:
        if (ValidInput(Num2)) {
          Num2 = Num2 + c;
          CurrentState = 51;
          PrintStatus();
          SpeakOut(c);
        }

        break;
      case 51:
        if ((c >= '0') && (c <= '9')) {
          if (ValidInput(Num2)) {
            Num2 = Num2 + c;
            PrintStatus();
            SpeakOut(c);
          }
        }
        if (c == 'E') {
          Num1 = Calculate(Num1, Num2, Operator).toString();
          CurrentState = 6;
          PrintStatus();
          SpeakOut(c);
        }

        break;
      case 6:
        if ((c >= '0') && (c <= '9')) {
          Num1 = c;
          Num2 = "";
          CurrentState = 1;
          PrintStatus();
          SpeakOut(c);
        }
        if ((c == '+') || (c == '-') || (c == '*') || (c == '/')) {
          Num2 = "";
          Operator = c;
          CurrentState = 3;
          PrintStatus();
          SpeakOut(c);
        }
        break;
      default:
        Serial.println("Input Error");
    }

    if (c == 'N') {
      CurrentState = 0;
      RecordStatus = 0xFF;
      Num1='0';
      Serial.println("Clear");
      PrintStatus();
      SpeakOut(c);
    }

    // If status change, we should send voice
    if (RecordStatus != CurrentState) {
      if (CurrentState == 6) {
        // 如果字符串后面有 "." 那么需要检查去掉多余的 0
        if (Num1.indexOf('.') != -1) {
          while (Num1.charAt(Num1.length() - 1) == '0') {
            Num1.remove(Num1.length() - 1);
          }
          if (Num1.charAt(Num1.length() - 1) == '.') {
            Num1.remove(Num1.length() - 1);
          }
        }
        LCDShow(Num1);

        for (int i = 0; i < Num1.length(); i++) {
          //Serial.print(Num1[i]);
          if (Num1[i] == '-') {
            SpeakOut('F');
          }
          else {
            SpeakOut(Num1[i]);
          }
          delay(500);
        }
      }
      
      if ((CurrentState == 0) || (CurrentState == 1) || (CurrentState == 2) || (CurrentState == 21)) {
        LCDShow(Num1);
      }
      if (CurrentState == 3) {
        LCDShow(Num1 + Operator);
      }
      if ((CurrentState == 4) || (CurrentState == 5) || (CurrentState == 51)) {
        LCDShow(Num1 + Operator + Num2);
      }

      RecordStatus == CurrentState;
    }
    c = 0;
  }
}
